#AutoIt3Wrapper_Au3Check_Parameters=-d -w 1 -w 2 -w 3 -w- 4 -w 5 -w 6 -w 7
#Tidy_Parameters=/sort_funcs /reel

#include <array.au3>
#include <date.au3>
#include "XMLWrapperEx.au3"

; This error hanlder will be used globally excepting inside UDF Functions
; unles you set it up for UDF with _XML_CreateObject()
Global $oErrorHandler = ObjEvent("AutoIt.Error", ErrFunc_CustomUserHandler_MAIN)
#forceref $oErrorHandler

Example_1__XML_SelectNodes()
Example_2__XML_CreateChildWAttr()
Example_3__XML_ErrorParser_GetDescription()
Example_4__XML_RemoveAttribute()
Example_5__XML_ReplaceChild()
Example_6__XML_GetChildNodes()
Example_7__XML_GetAllAttribIndex()
Example_MSDN_1__setAttributeNode()

#Region XMLWrapperEx__Examples.au3 - Function

Func Example_1__XML_SelectNodes()

	; first you must create $oXmlDoc object
	Local $oXmlDoc = _XML_CreateObject(Default, ErrFunc_CustomUserHandler_UDF)

	; now you can add EVENT Handler
	Local $oXMLDOM_EventsHandler = ObjEvent($oXmlDoc, "XML_DOM_EVENT_")
	#forceref $oXMLDOM_EventsHandler

	; Load file to $oXmlDoc
	Local $sXmlFile = @ScriptDir & "\XML_Example_XML_Files\MSDN\books.xml"
	_XML_Load($oXmlDoc, $sXmlFile)
	If @error Then MsgBox(0, '_XML_Load @error:', XML_My_ErrorParser(@error))

	; simple display $oXmlDoc - for checking only
;~ 	MsgBox($MB_SYSTEMMODAL + $MB_ICONINFORMATION, 'Example_1__XML_SelectNodes', _XML_TIDY($oXmlDoc))

	; selecting nodes
	Local $oNodesColl = _XML_SelectNodes($oXmlDoc, "//price")
	If @error Then MsgBox(0, '_XML_SelectNodes @error:', XML_My_ErrorParser(@error))

	; change Nodes Collection to array
	Local $aNodesColl = _XML_Array_GetNodesProperties($oNodesColl)

	; display array
	_ArrayDisplay($aNodesColl, 'Example_1__XML_SelectNodes : ' & 'Length=' & $oNodesColl.length & '   XPath=' & $oNodesColl.expr)

EndFunc   ;==>Example_1__XML_SelectNodes

Func Example_2__XML_CreateChildWAttr()

	; first you must create $oXmlDoc object
	Local $oXmlDoc = _XML_CreateObject(Default, ErrFunc_CustomUserHandler_UDF)

	; Load file to $oXmlDoc
	Local $sXmlFile = @ScriptDir & "\XML_Example_XML_Files\MSDN\books.xml"
	_XML_Load($oXmlDoc, $sXmlFile)

	; simple display $oXmlDoc - for checking only
	MsgBox($MB_SYSTEMMODAL + $MB_ICONINFORMATION, 'Before Example_2__XML_CreateChildWAttr()', _XML_TIDY($oXmlDoc))

	Local $aAttributeList[2][2] = [['Fist', '1'], ['Second', '2']]
	_XML_CreateChildWAttr($oXmlDoc, "//book", 'MyNewNode', $aAttributeList, "SomeText")

	; simple display $oXmlDoc - for checking only
	MsgBox($MB_SYSTEMMODAL + $MB_ICONINFORMATION, 'After Example_2__XML_CreateChildWAttr()', _XML_TIDY($oXmlDoc))

	$oXmlDoc.loadXML(_XML_TIDY($oXmlDoc))

	; View in default System Viewer
	_XML_Misc_Viewer($oXmlDoc)

EndFunc   ;==>Example_2__XML_CreateChildWAttr

Func Example_3__XML_ErrorParser_GetDescription()

	; first you must create $oXmlDoc object
	Local $oXmlDoc = _XML_CreateObject(Default, ErrFunc_CustomUserHandler_UDF)

	; Load file to $oXmlDoc
	Local $sXmlFile = @ScriptDir & "\XML_Example_XML_Files\MSDN\books.xml"
	_XML_Load($oXmlDoc, $sXmlFile)
	If @error = $XMLWRAPPER_ERROR_XPATH Then
		MsgBox($MB_SYSTEMMODAL + $MB_ICONINFORMATION, 'Test 1: ' & @error & ' ' & @extended, XML_My_ErrorParser(@error, @extended))
	EndIf

	; now we fires COM Error
	; selecting nodes with incorrect XPath expression
	Local $oNodesColl = _XML_SelectNodes($oXmlDoc, "/1/price")
	#forceref $oNodesColl
	If @error = $XMLWRAPPER_ERROR_XPATH Then
		MsgBox($MB_SYSTEMMODAL + $MB_ICONINFORMATION, 'Test 2: ' & @error & ' ' & @extended, XML_My_ErrorParser(@error, @extended))
		MsgBox($MB_SYSTEMMODAL + $MB_ICONINFORMATION, 'Test 3: _XML_ErrorParser_GetDescription', _XML_ErrorParser_GetDescription($oXmlDoc))
	EndIf

	; now we try fires $oXmlDoc.parseError.errorCode - wrong XML file
	; again create a new clean $oXmlDoc object
	$oXmlDoc = _XML_CreateObject()

	; Load file to $oXmlDoc
	$sXmlFile = @ScriptFullPath
	_XML_Load($oXmlDoc, $sXmlFile)
	If @error = $XMLWRAPPER_ERROR_PARSE Then
		MsgBox($MB_SYSTEMMODAL + $MB_ICONINFORMATION, 'Test 2: ' & @error & ' ' & @extended, XML_My_ErrorParser(@error, @extended))
		MsgBox($MB_SYSTEMMODAL + $MB_ICONINFORMATION, 'Test 3: _XML_ErrorParser_GetDescription', _XML_ErrorParser_GetDescription($oXmlDoc))
	EndIf

EndFunc   ;==>Example_3__XML_ErrorParser_GetDescription

Func Example_4__XML_RemoveAttribute()

	; first you must create $oXmlDoc object
	Local $oXmlDoc = _XML_CreateObject(Default, ErrFunc_CustomUserHandler_UDF)

	; Load file to $oXmlDoc
	Local $sXmlFile = @ScriptDir & "\XML_Example_XML_Files\MSDN\books.xml"
	_XML_Load($oXmlDoc, $sXmlFile)

	Local $aAttributeList[2][2] = [['Fist', '1'], ['Second', '2']]
	_XML_CreateChildWAttr($oXmlDoc, "//book", 'MyNewNode', $aAttributeList, "SomeText")

	; simple display $oXmlDoc - for checking only
	MsgBox($MB_SYSTEMMODAL + $MB_ICONINFORMATION, 'After Example_2__XML_CreateChildWAttr()', _XML_TIDY($oXmlDoc))

	_XML_RemoveAttribute($oXmlDoc, "//MyNewNode[1]", 'Second')

	; TIDY xml and save result back to object
	$oXmlDoc.loadXML(_XML_TIDY($oXmlDoc))

	; View in default System Viewer
	_XML_Misc_Viewer($oXmlDoc)

EndFunc   ;==>Example_4__XML_RemoveAttribute

Func Example_5__XML_ReplaceChild()

	; first you must create $oXmlDoc object
	Local $oXmlDoc = _XML_CreateObject(Default, ErrFunc_CustomUserHandler_UDF)

	; Load file to $oXmlDoc
	Local $sXmlFile = @ScriptDir & "\XML_Example_XML_Files\MSDN\books.xml"
	_XML_Load($oXmlDoc, $sXmlFile)

	_XML_ReplaceChild($oXmlDoc, '//price', 'TextBook')

	; TIDY xml and save result back to object
	$oXmlDoc.loadXML(_XML_TIDY($oXmlDoc))

	; View in default System Viewer
	_XML_Misc_Viewer($oXmlDoc)

EndFunc   ;==>Example_5__XML_ReplaceChild

Func Example_6__XML_GetChildNodes()

	; first you must create $oXmlDoc object
	Local $oXmlDoc = _XML_CreateObject(Default, ErrFunc_CustomUserHandler_UDF)

	; Load file to $oXmlDoc
	Local $sXmlFile = @ScriptDir & "\XML_Example_XML_Files\MSDN\books.xml"
	_XML_Load($oXmlDoc, $sXmlFile)

	; simple display $oXmlDoc - for checking only
	MsgBox($MB_SYSTEMMODAL + $MB_ICONINFORMATION, 'Example_6__XML_GetChildNodes', _XML_TIDY($oXmlDoc))

	; selecting nodes
	Local $oNodesColl = _XML_GetChildNodes($oXmlDoc, "//book")

	; change Nodes Collection to array
	Local $aNodesColl = _XML_Array_GetNodesProperties($oNodesColl)
	_ArrayDisplay($aNodesColl, 'Example_6__XML_GetChildNodes: ' & 'Length=' & $oNodesColl.length & '   XPath=' & $oNodesColl.expr)

EndFunc   ;==>Example_6__XML_GetChildNodes

Func Example_7__XML_GetAllAttribIndex()

	; first you must create $oXmlDoc object
	Local $oXmlDoc = _XML_CreateObject(Default, ErrFunc_CustomUserHandler_UDF)

	; Load file to $oXmlDoc
	Local $sXmlFile = @ScriptDir & "\XML_Example_XML_Files\MSDN\books.xml"
	_XML_Load($oXmlDoc, $sXmlFile)

	Local $aAttributeList[2][2] = [['Fist', '1'], ['Second', '2']]
	_XML_CreateChildWAttr($oXmlDoc, "//book", 'MyNewNode', $aAttributeList, "SomeText")

	_XML_RemoveAttribute($oXmlDoc, "//MyNewNode[1]", 'Second')

	; TIDY xml and save result back to object
	$oXmlDoc.loadXML(_XML_TIDY($oXmlDoc))

	; View in default System Viewer
	_XML_Misc_Viewer($oXmlDoc)

	Local $oAttriubtes = _XML_GetAllAttribIndex($oXmlDoc, '//MyNewNode', 0)
	Local $aAttributesList = _XML_Array_GetAttributesProperties($oAttriubtes)

	_ArrayDisplay($aAttributesList, '$aAttributesList')

EndFunc   ;==>Example_7__XML_GetAllAttribIndex

Func Example_MSDN_1__setAttributeNode()

	; first you must create $oXmlDoc object
	Local $oXmlDoc = _XML_CreateObject(Default, ErrFunc_CustomUserHandler_UDF)

	; Load file to $oXmlDoc
	Local $sXmlFile = @ScriptDir & "\XML_Example_XML_Files\MSDN\books.xml"
	_XML_Load($oXmlDoc, $sXmlFile)

	; simple display $oXmlDoc - for checking only
	MsgBox($MB_SYSTEMMODAL + $MB_ICONINFORMATION, 'Example_MSDN_1__setAttributeNode', _XML_TIDY($oXmlDoc))

	;MSDN; https://msdn.microsoft.com/en-us/library/ms762235(v=vs.85).aspx

	;MSDN; nodePublishDate = xmlDoc.createAttribute("PublishDate");
	Local $oNodePublishDate = $oXmlDoc.createAttribute("PublishDate");

	; simple display $oXmlDoc - for checking only
	MsgBox($MB_SYSTEMMODAL + $MB_ICONINFORMATION, 'Temp', _XML_TIDY($oXmlDoc)) ; Hmmm... Nothing happend

	;MSDN; nodePublishDate.value = String(Date());
	$oNodePublishDate.value = String(_Now());

	;MSDN; nodeBook = xmlDoc.selectSingleNode("//book");
	Local $oNodeBook = $oXmlDoc.selectSingleNode("//book");

	;MSDN; nodeBook.setAttributeNode(nodePublishDate);
	$oNodeBook.setAttributeNode($oNodePublishDate);

	;MSDN; WScript.Echo(nodeBook.getAttribute("PublishDate"));
	ConsoleWrite($oNodeBook.getAttribute("PublishDate") & @CRLF)

	; simple display $oXmlDoc - for checking only
	MsgBox($MB_SYSTEMMODAL + $MB_ICONINFORMATION, 'Result', _XML_TIDY($oXmlDoc))

	; this following part is not related to MSDN but is usefull to show how _XML_Array_GetNodesProperties display attributes
	Local $oNodesColl = _XML_SelectNodes($oXmlDoc, "//book")

	; change Nodes Collection to array
	Local $aNodesColl = _XML_Array_GetNodesProperties($oNodesColl)
	_ArrayDisplay($aNodesColl, 'Example_1__XML_SelectNodes : ' & $oNodesColl.expr)

EndFunc   ;==>Example_MSDN_1__setAttributeNode
#EndRegion XMLWrapperEx__Examples.au3 - Function

#Region XMLWrapperEx__Examples.au3 - XML DOM Error/Event Handling

Func ErrFunc_CustomUserHandler_MAIN($oError)
	ConsoleWrite(@ScriptName & " (" & $oError.scriptline & ") : MainScript ==> COM Error intercepted !" & @CRLF & _
			@TAB & "err.number is: " & @TAB & @TAB & "0x" & Hex($oError.number) & @CRLF & _
			@TAB & "err.windescription:" & @TAB & $oError.windescription & @CRLF & _
			@TAB & "err.description is: " & @TAB & $oError.description & @CRLF & _
			@TAB & "err.source is: " & @TAB & @TAB & $oError.source & @CRLF & _
			@TAB & "err.helpfile is: " & @TAB & $oError.helpfile & @CRLF & _
			@TAB & "err.helpcontext is: " & @TAB & $oError.helpcontext & @CRLF & _
			@TAB & "err.lastdllerror is: " & @TAB & $oError.lastdllerror & @CRLF & _
			@TAB & "err.scriptline is: " & @TAB & $oError.scriptline & @CRLF & _
			@TAB & "err.retcode is: " & @TAB & "0x" & Hex($oError.retcode) & @CRLF & @CRLF)
EndFunc   ;==>ErrFunc_CustomUserHandler_MAIN

Func ErrFunc_CustomUserHandler_UDF($oError)
	ConsoleWrite(@ScriptDir & '\XMLWrapperEx.au3' & " (" & $oError.scriptline & ") : UDF ==> COM Error intercepted ! " & @CRLF & _
			@TAB & "err.number is: " & @TAB & @TAB & "0x" & Hex($oError.number) & @CRLF & _
			@TAB & "err.windescription:" & @TAB & $oError.windescription & @CRLF & _
			@TAB & "err.description is: " & @TAB & $oError.description & @CRLF & _
			@TAB & "err.source is: " & @TAB & @TAB & $oError.source & @CRLF & _
			@TAB & "err.helpfile is: " & @TAB & $oError.helpfile & @CRLF & _
			@TAB & "err.helpcontext is: " & @TAB & $oError.helpcontext & @CRLF & _
			@TAB & "err.lastdllerror is: " & @TAB & $oError.lastdllerror & @CRLF & _
			@TAB & "err.scriptline is: " & @TAB & $oError.scriptline & @CRLF & _
			@TAB & "err.retcode is: " & @TAB & "0x" & Hex($oError.retcode) & @CRLF & @CRLF)
EndFunc   ;==>ErrFunc_CustomUserHandler_UDF

Func XML_DOM_EVENT_ondataavailable()
	#CS
		ondataavailable Event
		https://msdn.microsoft.com/en-us/library/ms754530(v=vs.85).aspx
	#CE
	Local $oXmlDoc = @COM_EventObj
	#forceref $oXmlDoc
	Local $sMessage = 'XML_DOM_EVENT_ fired "ondataavailable"' & @CRLF
	ConsoleWrite($sMessage)
EndFunc   ;==>XML_DOM_EVENT_ondataavailable

Func XML_DOM_EVENT_onreadystatechange()
	#CS
		onreadystatechange Event
		https://msdn.microsoft.com/en-us/library/ms759186(v=vs.85).aspx
	#CE
	Local $oXmlDoc = @COM_EventObj
	#forceref $oXmlDoc
	Local $sMessage = 'XML_DOM_EVENT_ fired "onreadystatechange" : ReadyState = ' & $oXmlDoc.ReadyState & @CRLF
	ConsoleWrite($sMessage)

EndFunc   ;==>XML_DOM_EVENT_onreadystatechange

Func XML_DOM_EVENT_ontransformnode($oNodeCode_XSL, $oNodeData_XML, $bBool)
	#forceref $oNodeCode_XSL, $oNodeData_XML, $bBool
	#CS
		ontransformnode Event
		https://msdn.microsoft.com/en-us/library/ms767521(v=vs.85).aspx
	#CE
	Local $oXmlDoc = @COM_EventObj
	#forceref $oXmlDoc
	Local $sMessage = 'XML_DOM_EVENT_ fired "ontransformnode"' & @CRLF
	ConsoleWrite($sMessage)

EndFunc   ;==>XML_DOM_EVENT_ontransformnode
; #FUNCTION# ====================================================================================================================
; Name ..........: XML_My_ErrorParser
; Description ...: Changing $XMLWRAPPER_ERROR_... to human readable description
; Syntax ........: XML_My_ErrorParser($iXMLWrapper_Error, $iExtended)
; Parameters ....: $iXMLWrapper_Error	- an integer value.
;                  $iExtended           - an integer value.
; Return values .: None
; Author ........: mLipok
; Modified ......:
; Remarks .......:
; Related .......:
; Link ..........:
; Example .......: No
; ===============================================================================================================================
Func XML_My_ErrorParser($iXMLWrapper_Error, $iExtended = 0)
	Switch $iXMLWrapper_Error
		Case $XMLWRAPPER_ERROR_OK
			Return 'All is ok.'
		Case $XMLWRAPPER_ERROR_GENERAL
			Return 'The error which is not specifically defined.'
		Case $XMLWRAPPER_ERROR_ISNOTOBJECT
			Return 'No object passed to function'
		Case $XMLWRAPPER_ERROR_ISNOTVALIDOBJECT
			Return 'Invalid object passed to function'
		Case $XMLWRAPPER_ERROR_PARSE
			Return 'Error: with Parsing objects, .parseError.errorCode=' & $iExtended
		Case $XMLWRAPPER_ERROR_LOAD
			Return 'Error opening specified file.'
		Case $XMLWRAPPER_ERROR_SAVE
			Return 'Error saving file.'
		Case $XMLWRAPPER_ERROR_SAVEFILERO
			Return 'Destination file is Read Only.'
		Case $XMLWRAPPER_ERROR_XPATH
			Return 'XPath syntax error - check also COM Error Handler.'
		Case $XMLWRAPPER_ERROR_DOMVERSION
			Return 'DOM Version: ' & 'MSXML Version ' & $iExtended & ' or greater required for this function'
		Case $XMLWRAPPER_ERROR_PARAMETER
			Return 'Wrong parameter passed to function.'
		Case $XMLWRAPPER_ERROR_ARRAY
			Return 'Wrong array parameter passed to function. Check array dimension and conent.'
		Case $XMLWRAPPER_ERROR_NODECREATEAPPEND
			Return 'Can not create or append Node'
		Case $XMLWRAPPER_ERROR_EMPTYCOLLECTION
			Return 'Collections of objects was empty'
		Case $XMLWRAPPER_ERROR_NONODESMATCH
			Return 'No nodes match the XPath expression'
		Case $XMLWRAPPER_ERROR_PARSE_XSL
			Return 'Error with Parsing XSL objects use _XML_ErrorParser_GetDescription() for Get details'
		Case $XMLWRAPPER_ERROR_NOATTRMATCH
			Return 'There is no such attribute in selected node.'
		Case $XMLWRAPPER_ERROR_NOCHILDMATCH
			Return 'There is no Child in nodes matched by XPath expression.'
		Case $XMLWRAPPER_ERROR_OJBCREATE
			Return 'Object can not be created.'
		Case Else
			Return 'NO ERROR DESCRIPTION FOR:' & $iXMLWrapper_Error
	EndSwitch
EndFunc   ;==>XML_My_ErrorParser
#EndRegion XMLWrapperEx__Examples.au3 - XML DOM Error/Event Handling

#Region XMLWrapperEx__Examples.au3 - Help, Documentation, Manuals
#CS
	A Beginner's Guide to the XML DOM
	https://msdn.microsoft.com/en-us/library/aa468547.aspx

	DOM Reference
	https://msdn.microsoft.com/en-us/library/ms764730(v=vs.85).aspx

	XML DOM Properties
	https://msdn.microsoft.com/en-us/library/ms763798(v=vs.85).aspx

	XML DOM Methods
	https://msdn.microsoft.com/en-us/library/ms757828(v=vs.85).aspx

	XML Glossary
	https://msdn.microsoft.com/en-us/library/ms256452(v=vs.85).aspx

	MSXML API History
	https://msdn.microsoft.com/en-us/library/ms762314(v=vs.85).aspx

	XML Standards Reference
	https://msdn.microsoft.com/en-us/library/ms256177(v=vs.85).aspx

	XSLT Reference
	https://msdn.microsoft.com/en-us/library/ms256069(v=vs.85).aspx

	XPath Reference
	https://msdn.microsoft.com/en-us/library/ms256115(v=vs.85).aspx

	XPath Syntax
	https://msdn.microsoft.com/en-us/library/ms256471(v=vs.85).aspx

	XPath Examples
	https://msdn.microsoft.com/en-us/library/ms256086(v=vs.110).aspx

	Location Path Examples
	https://msdn.microsoft.com/en-us/library/ms256236(v=vs.110).aspx

	Microsoft XML DOM output has no CR/LF
	http://www.experts-exchange.com/Web_Development/Web_Languages-Standards/Q_20135543.html

	Formatting of XML file
	http://www.visualbasicscript.com/Formatting-of-XML-file-m77414.aspx

	Stream Object (ADO)
	https://msdn.microsoft.com/en-us/library/windows/desktop/ms675032(v=vs.85).aspx

	TODO:
	IXMLDOMAttribute Members  ($oAttribute)
	https://msdn.microsoft.com/en-us/library/ms767677(v=vs.85).aspx
#CE
#EndRegion XMLWrapperEx.au3 - Help, Documentation, Manuals

#Region XMLWrapperEx__Examples.au3 - TOOLs

#CS ON LINE XPath Tools
	http://www.xpathtester.com/xpath
	http://codebeautify.org/Xpath-Tester
	http://xpath.online-toolz.com/tools/xpath-editor.php
	http://www.whitebeam.org/library/guide/TechNotes/xpathtestbed.rhtm
	http://www.qutoric.com/xslt/analyser/xpathtool.html
	https://addons.mozilla.org/en-US/firefox/addon/xpather/
	http://www.xmlme.com/XpathTool.aspx
#CE

#CS OFF LINE XPath Tools
	http://qutoric.com/xmlquire/
	https://xpathvisualizer.codeplex.com/
#CE

#EndRegion XMLWrapperEx.au3 - TOOLs
